#version 450 core

layout (location = 0) out vec4 FragColor;

in vec3 v_Normal;
in vec2 v_TexCoord;
in vec4 v_Vertex;
in vec4 v_LightSpacePos;

layout (location = 4) uniform vec4  u_LightPosition;
layout (location = 5) uniform float u_zOffset;
layout (binding = 0) uniform sampler2D       u_SceneTexture;
layout (binding = 1) uniform sampler2D       u_ShadowTexture;

void main() {
    // Compute fragment diffuse color
    vec3 N      = normalize(v_Normal);
    vec3 L      = normalize(u_LightPosition.xyz - v_Vertex.xyz);
    float NdotL = max(dot(N, L), 0.0);
    vec4 color  = texture(u_SceneTexture, v_TexCoord) * NdotL;

    // Compute shadow factor (if statements avoided)
    vec4 shadow = vec4(1.0);
    // vertex in lightspace has to undergo persp. division and ndc transform
    vec3 coord = v_LightSpacePos.xyz / v_LightSpacePos.w;
    coord = coord * 0.5 + 0.5;
    float shadowMapValue = texture(u_ShadowTexture, coord.xy).r;
    // use user provided offset value
    shadow -= vec4(shadowMapValue < (coord.z - u_zOffset));

    // Modulate fragment's color according to result of shadow test
    FragColor = color*shadow;
}
