#version 450 core

layout (location = 0) out vec4 FragColor;

layout (binding = 0)  uniform sampler2D u_zMap;
layout (r32ui, binding = 1) uniform uimage2D u_importanceMap;

layout (location = 0) uniform mat4 u_inversePVM;
layout (location = 1) uniform mat4 u_lightPVM;
layout (location = 2) uniform vec2 u_viewport;
layout (location = 3) uniform float u_resolution;
layout (location = 4) uniform float u_floatScale;

void main(void) {
    ivec2 pixelCoord = ivec2(gl_FragCoord.xy);
    float z = texelFetch(u_zMap, pixelCoord, 0).r;
    vec4 projCoord = vec4(vec2(pixelCoord.xy), z, 1.0);
    projCoord.xy = projCoord.xy / u_viewport.xy;

    // into world space
    projCoord.xyz = projCoord.xyz * 2.0 - 1.0;
    projCoord = u_inversePVM * projCoord;

    // into light space
    projCoord = u_lightPVM * projCoord;
	projCoord.xyz = projCoord.xyz / projCoord.w;
    projCoord.xyz = projCoord.xyz * 0.5 + 0.5;

	uint imp = uint(u_floatScale - (u_floatScale * z));
	ivec2 desiredPos = ivec2(projCoord.xy * u_resolution);

    // closest valid
    imageAtomicMax(u_importanceMap, desiredPos, imp); 
}
