#version 450 core

layout (location = 0) out vec4 FragColor;

layout (r32ui, binding = 0) uniform uimage2D u_importanceMap;

layout (location = 1) uniform mat4 u_inverseLightPVM;
layout (location = 2) uniform mat4 u_cameraPVM;
layout (location = 3) uniform vec2 u_viewport;
layout (location = 4) uniform float u_resolution;
layout (location = 5) uniform float u_floatScale;

void main(void) {
    // depth map
    FragColor = vec4(gl_FragCoord.z, 0.0, 0.0, 1.0);

    ivec2 pixelCoord = ivec2(gl_FragCoord.xy);
    vec4 projCoord = vec4(gl_FragCoord.xyz, 1.0);
    projCoord.xy = projCoord.xy / u_resolution;

    // into world space
    projCoord.xyz = projCoord.xyz * 2.0 - 1.0;
    projCoord = u_inverseLightPVM * projCoord;

    // into light space
    projCoord = u_cameraPVM * projCoord;
	projCoord.xyz = projCoord.xyz / projCoord.w;
    projCoord.xyz = projCoord.xyz * 0.5 + 0.5;

    // desired view function
    if (projCoord.x > 1.0 || projCoord.x < 0.0 || projCoord.y > 1.0 || projCoord.y < 0.0)
    {
        return;
    }

	uint imp = uint(u_floatScale - (u_floatScale * projCoord.z));

    imageAtomicMax(u_importanceMap, pixelCoord, imp); 
}
