#version 450 core

layout (location = 0) out vec4 FragColor;

layout (binding = 0)  uniform sampler2D u_render;
layout (binding = 1)  uniform sampler2D u_zRender;
layout (binding = 2)  uniform sampler2D u_warpMapX;
layout (binding = 3)  uniform sampler2D u_warpMapY;
layout (binding = 4)  uniform sampler2D u_shadowMap;

layout (location = 0) uniform mat4 u_inversePVM;
layout (location = 1) uniform mat4 u_lightPVM;
layout (location = 2) uniform float u_zOffset;
layout (location = 3) uniform vec2 u_viewport;

void main(void) {
    ivec2 pixelCoord = ivec2(gl_FragCoord.xy);
    vec4 color = vec4(texelFetch(u_render, pixelCoord, 0).rgb, 1.0);
    float z = texelFetch(u_zRender, pixelCoord, 0).r;
    if (z == 1.0)
    {
        FragColor.xyz = color.xyz;
        return;
    }
    vec4 projCoord = vec4(pixelCoord.xy, z, 1.0);
    projCoord.xy = projCoord.xy / u_viewport.xy;


    // into world space
    projCoord.xyz = projCoord.xyz * 2.0 - 1.0;
    projCoord = u_inversePVM * projCoord;

    // into light space
    projCoord = u_lightPVM * projCoord;
	projCoord.xyz = projCoord.xyz / projCoord.w;
    projCoord.xyz = projCoord.xyz * 0.5 + 0.5;

    // warp shadow map coords
    float valX = texture(u_warpMapX, vec2(projCoord.x, 0.5)).r - texture(u_warpMapX, vec2(projCoord.x, 0.5)).g;
    float valY = texture(u_warpMapY, vec2(projCoord.y, 0.5)).r - texture(u_warpMapY, vec2(projCoord.y, 0.5)).g;

    vec2 uvs = vec2(projCoord.x + valX, projCoord.y + valY);
    float shadowZ = texture(u_shadowMap, uvs).r;

    float shadow = 1.0;
    shadow = shadow - float(shadowZ < (projCoord.z - u_zOffset));

    // Modulate fragment's color according to result of shadow test
    FragColor.xyz = color.xyz * shadow;

    
}