#version 450 core

#ifdef USER_TEST
#endif

layout (location = 0) out vec4 FragColor;

in vec3 v_Normal;
in vec2 v_TexCoord;
in vec4 v_Vertex;
in vec4 v_LightSpacePos;

layout (binding = 0) uniform sampler2D       u_SceneTexture;
layout (binding = 1) uniform sampler2D       u_ShadowTexture;
layout (binding = 2) uniform sampler2D       u_warpMapX;
layout (binding = 3) uniform sampler2D       u_warpMapY;

layout (location = 4) uniform float u_zOffset;
layout (location = 5) uniform vec4  u_LightPosition;

void main() {
    // Compute fragment diffuse color
    vec3 N      = normalize(v_Normal);
    vec3 L      = normalize(u_LightPosition.xyz - v_Vertex.xyz);
    float NdotL = max(dot(N, L), 0.0);
    vec4 color  = texture(u_SceneTexture, v_TexCoord) * NdotL;

    // Compute shadow factor (if statements avoided)
    vec4 shadow = vec4(1.0);
    // vertex in lightspace has to undergo persp. division and ndc transform
    vec3 coord = v_LightSpacePos.xyz / v_LightSpacePos.w;
    coord = coord * 0.5 + 0.5;

    // warp shadow map coords
    float valX = texture(u_warpMapX, vec2(coord.x, 0.5)).r - texture(u_warpMapX, vec2(coord.x, 0.5)).g;
    float valY = texture(u_warpMapY, vec2(coord.y, 0.5)).r - texture(u_warpMapY, vec2(coord.y, 0.5)).g;
    vec2 uvs = vec2(coord.x + valX, coord.y + valY);

    float shadowMapValue = texture(u_ShadowTexture, uvs.xy).r;
    // use user provided offset value
    shadow -= vec4(shadowMapValue < (coord.z - u_zOffset));

    // Modulate fragment's color according to result of shadow test
    FragColor = color*shadow;
}
